/*
Align stereo images so their centers don't move.

 Orion Sky Lawlor, olawlor@acm.org, 2007-04-02
*/
#include <stdio.h>
#include <stdlib.h>
#include "osl/osl.h"
#include "osl/raster.h"
#include "osl/movie.h"
#include "osl/gui.h"

#include <complex>
#include <fftw3.h>  /* Download from www.fftw.org */
// Floating-point complex type
typedef std::complex<float> fcpx;

using namespace osl;
using namespace osl::graphics2d;
using namespace osl::gui;

#define outdec 1 /* Decimation factor to apply to displayed image */

class CpxRaster:public FlatRasterT<fcpx> {
	typedef FlatRasterT<fcpx> super;
public:
	FlatRasterConstructors(CpxRaster)

	virtual unsigned int getProperties(void) const;

	//Single-Pixel interface (must be in-bounds)
	virtual Color getColor(int x,int y) const;
	virtual void setColor(int x,int y,const Color &c);
};


unsigned int CpxRaster::getProperties(void) const
{
	return COLOR_23bit|CONTIGUOUS_ROWS;
}
 
//Single-Pixel interface (must be in-bounds)
Color CpxRaster::getColor(int x,int y) const {
	fcpx c=at(x,y);
	double scl=1.0;
	return Color(c.real()*scl,c.imag()*scl,0.0);
}
void CpxRaster::setColor(int x,int y,const Color &c) {
	at(x,y)=fcpx(c.r,c.g);
}


class offset_finder {
public:
	/** Allocate memory for finding offsets in images of this size. */
	offset_finder(const Rect &size);
	
	/** Return the origin of raster B in raster A.  
	    B must be smaller than A.  */
	Vector2d find_offset(const RgbaRaster &A,const RgbaRaster &B,float *confidence=0);
	
	/** Extract the correlation image from the last offset search. */
	void last_corr(const RgbaRaster &A,const RgbaRaster &B,RgbaRaster &corr);
	

private:
	int fftW,fftH; /* size of FFTs to perform */
	fftwf_plan A_fwfft;
	fftwf_plan B_fwfft;
	fftwf_plan B_bwfft;
	CpxRaster cA, cB; /* complex rasters */
	float corr_strength(int x,int y) {
		if (x<0) x+=fftW; if (y<0) y+=fftH;
		if (x>=fftW) x-=fftW; if (y>=fftH) y-=fftH;
		return cB.at(x,y).real();
	}
	void fill_corr(const RgbaRaster &A,const RgbaRaster &B);
	void copyRGtoCpx(const RgbaRaster &src,CpxRaster &dest);
	void copyBAtoCpx(const RgbaRaster &src,CpxRaster &dest);
	void copyCpxtoRG(const CpxRaster &src,RgbaRaster &dest,float scl=1.0);
	/* Take the conjugate product of these two complex images. */
	void conj_prod(const CpxRaster &src,CpxRaster &dest);
};

/* pixels to fade in image around boundary */
enum {fade_in=50};
/* Fade-in function */
float *make_fade(void) {
	float *fade=new float[fade_in];
	for (int x=0;x<fade_in;x++)
		fade[x]=0.5*(1.0-cos((x+0.5f)*(M_PI/fade_in)));
	return fade;
}
void offset_finder::copyRGtoCpx(const RgbaRaster &src,CpxRaster &dest)
{
	//printf("%d x %d,  %d x %d\n",src.wid,src.ht,dest.wid,dest.ht);
	vassert(src.wid<=dest.wid,"Source image width must be less than dest");
	vassert(src.ht<=dest.ht,"Source image height must be less than dest");
	double scl=(1.0/256)*(1.0/(src.wid*src.ht)); /* normalizing scale factor */
	static float *fade=make_fade();
	int x,y;
	for (y=0;y<src.ht;y++) {
		float yscl=scl;
		if (y<fade_in) yscl*=fade[y];
		if (src.ht-y<fade_in) yscl*=fade[src.ht-y];
		for (x=0;x<src.wid;x++) {
			float xscl=1.0;
			if (x<fade_in) xscl*=fade[x];
			if (src.wid-x<fade_in) xscl*=fade[src.wid-x];
			RgbaPixel p=src.at(x,y);
			dest.at(x,y)=fcpx(p.r()*xscl*yscl,p.g()*xscl*yscl);
		}
		/* Fill remainder of row with zeros */
		for (;x<dest.wid;x++) dest.at(x,y)=0;
	}
	/* Fill remainder of destination with zeros */
	for (;y<dest.ht;y++) for (x=0;x<dest.wid;x++) dest.at(x,y)=0;
}
void offset_finder::copyBAtoCpx(const RgbaRaster &src,CpxRaster &dest)
{
	double scl=(1.0/256)*(1.0/(src.wid*src.ht)); /* normalizing scale factor */
	int x,y;
	for (y=0;y<src.ht;y++) {
		for (x=0;x<src.wid;x++) {
			RgbaPixel p=src.at(x,y);
			dest.at(x,y)=fcpx(p.b()*scl,p.a()*scl);
		}
		/* Fill remainder of row with zeros */
		for (;x<dest.wid;x++) dest.at(x,y)=0;
	}
	/* Fill remainder of destination with zeros */
	for (;y<dest.ht;y++) for (x=0;x<dest.wid;x++) dest.at(x,y)=0;
}

unsigned char toByte(float f) {
	if (f<0) return 0;
	if (f>1) return 255;
	else return (unsigned char)(f*255.0);
}

void offset_finder::copyCpxtoRG(const CpxRaster &src,RgbaRaster &dest,float scl) {
	int x,y;
	for (y=0;y<dest.ht;y++) for (x=0;x<dest.wid;x++) {
		fcpx s=src.at(x,y);
		dest.at(x,y)=RgbaPixel(toByte(scl*s.real()),toByte(scl*s.imag()),0);
	}
}

void offset_finder::conj_prod(const CpxRaster &src,CpxRaster &dest)
{
	int x,y;
	int r_thresh=10; /* low frequencies to zero out */
	int r2_thresh=r_thresh*r_thresh;
	int r2_threshhi=r2_thresh+200;
	for (y=0;y<dest.ht;y++) for (x=0;x<dest.wid;x++) {
		fcpx a=src.at(x,y);
		fcpx b=dest.at(x,y);
		fcpx c=a*std::conj(b);

		if (1) 
		{/* zero out low-freqency terms (edge detect) */
			int fx=x; if (fx>(fftW/2)) fx-=fftW;
			int fy=y; if (fy>(fftH/2)) fy-=fftH;
			int r2=(fx*fx+fy*fy);
			if (r2<=r2_thresh) c=0; 
			if (r2<=r2_threshhi) {
				c=c*((float)r2-r2_thresh)*(1.0f/(r2_threshhi-r2_thresh));
			}
		}
		dest.at(x,y)=40.0f*c;
	}
}

/** Fill the cB image with the correlation of these two images */
void offset_finder::fill_corr(const RgbaRaster &A,const RgbaRaster &B)
{
	// Compute correlation image for Red and Green channels
	copyRGtoCpx(A,cA); fftwf_execute(A_fwfft);
	copyRGtoCpx(B,cB); fftwf_execute(B_fwfft);
	conj_prod(cA,cB); fftwf_execute(B_bwfft);
	
	//FIXME: add in blue channel too?
}


/** Return the correlation image of raster B in raster A. */
void offset_finder::last_corr(const RgbaRaster &A,const RgbaRaster &B,RgbaRaster &corr)
{
	corr.reallocate(A.wid-B.wid, A.ht-B.ht);
	// corr.reallocate(A.wid, A.ht); // for debugging
	copyCpxtoRG(cB,corr);
}

/** Return the distance from the start of A to the start of B */
Vector2d offset_finder::find_offset(const RgbaRaster &A,const RgbaRaster &B,float *confidence)
{
	fill_corr(A,B);
	
	/* Loop over the correlation image to find the biggest correlation */
	Vector2d off(-999);
	float mag=0.0;
	int x,y;
	for (y=0;y<A.ht-B.ht;y++) 
	for (x=0;x<A.wid-B.wid;x++) {
		float m=corr_strength(x,y);
		if (m>mag) { /* New biggest correlation! */
			mag=m;
			off=Vector2d(x,y);
		}
	}
	if (off==Vector2d(-999)) {
		if (confidence) *confidence=0;
		return Vector2d(0,0);
	}
	
	/* Optimize to get sub-pixel precision:
	   Perform parabolic interpolation on peak.*/
	x=(int)off.x; y=(int)off.y;
	float a,b,c,d;
	a=corr_strength(x-1,y);
	b=corr_strength(x  ,y);
	c=corr_strength(x+1,y);
	d=4*((a+c)/2-b);
	if (d!=0)
		off.x=x+(a-c)/d;
	a=corr_strength(x,y-1);
	b=corr_strength(x,y  );
	c=corr_strength(x,y+1);
	d=4*((a+c)/2-b);
	if (d!=0)
	        off.y=y+(a-c)/d;
	
	/* FIXME: confidence-check */
	
	return off;
	
}

offset_finder::offset_finder(const Rect &size)
	:
	fftW(osl::roundUp2(size.getWidth())),
	fftH(osl::roundUp2(size.getHeight())),
	cA(fftW,fftH), cB(fftW,fftH)
{
	int flags=FFTW_DESTROY_INPUT + FFTW_ESTIMATE; 
	/* Quickly guess how to do FFT 
	WARNING: Without FFTW_ESTIMATE:
		- making a plan is much slower
		- making a plan trashes both input and output arrays
		- The FFT's may be faster
	*/
	A_fwfft=fftwf_plan_dft_2d(
		fftH,fftW,(fftwf_complex *)&cA.at(0,0),
		(fftwf_complex *)&cA.at(0,0),FFTW_FORWARD,flags);
	B_fwfft=fftwf_plan_dft_2d(
		fftH,fftW,(fftwf_complex *)&cB.at(0,0),
		(fftwf_complex *)&cB.at(0,0),FFTW_FORWARD,flags);
	B_bwfft=fftwf_plan_dft_2d(
		fftH,fftW,(fftwf_complex *)&cB.at(0,0),
		(fftwf_complex *)&cB.at(0,0),FFTW_BACKWARD,flags);
}


int main(int argc,char *argv[])
{
	if (argc<3) {
		printf("Usage: fft_stereo <left image> <right image>\n");
	}
	const char *fL=argv[1], *fR=argv[2]; /* filenames: left and right */
	
	Vector2d off;
	{
		/* Load in minified images */
		RgbaRaster L,R;
		int shrink=4; /* scales a 1600-pixel across image down to below 400 pixels */
		MiniRaster sL(shrink,L), sR(shrink,R);
		sL.readNoThrow(fL); sR.readNoThrow(fR);

		/* Find the gross offset between minfied images */
		offset_finder f(L);
		Vector2d subRstart(R.wid/4,R.ht/4); /* corner of R to extract */
		RgbaRaster subR(R.wid/2,R.ht/2, R, 
			(int)subRstart.x, (int)subRstart.y);
		off=(f.find_offset(L,subR)-subRstart)*shrink;

		if (1) { /* write out correlation image */
			RgbaRaster corr;
			f.last_corr(L,subR,corr);
			corr.writeNoThrow("corr.ppm");
		}
	}
	printf("Right-image offset: %.1f x %.1f pixels\n", off.x,off.y);
	
	/* Resample full-res images to offset */
	RgbaRaster L,R;
	L.readNoThrow(fL); R.readNoThrow(fR);
	RgbaRaster outRast(L.wid,L.ht); Rasterizer out(outRast);
	out.clear(Color::black);
	GraphicsState gs; gs.translate(off.x,off.y);
	out.copy(gs,R);
	
	outRast.writeNoThrow("out.ppm");
	R.readNoThrow("out.ppm"); /* Silly hack-- R=outRast, resample right image */
	
	/* Search in a grid over the images for stereo offsets */
	int lW=osl::max(32,osl::roundUp2(L.wid/15)), lH=osl::roundUp2(L.ht/30); /* half size of L chip */
	printf("Searching on grid of %d x %d chips...\n",lW,lH);
	int rW=lW/2, rH=lH/2; /* half size of R chip */
	int gW=rW/2, gH=rH/2; /* grid size */
	offset_finder f(Rect(2*lW,2*lH));
	
	FILE *offsets=fopen("offsets.txt","w");
	for (int gY=lH, gYc=0; gY<=L.ht -lH; gY+=gH,gYc++)
	for (int gX=lW, gXc=0; gX<=L.wid-lW; gX+=gW,gXc++)
	{
		RgbaRaster subL(2*lW,2*lH, L, gX-lW,gY-lH);
		RgbaRaster subR(2*rW,2*rH, R, gX-rW,gY-rH);
		float confidence=1.0;
		Vector2d off=f.find_offset(subL,subR,&confidence)-Vector2d(lW-rW,lH-rH);
		GraphicsState gs; gs.setLineWidth(rW*0.03); gs.setColor(Color::green);
		Vector2d c(gX,gY);
		out.lineSegment(gs, c, c+off);
		out.fill(gs,CircleShape(c,rW*0.05));
		fprintf(offsets,"%d\t%d\t%.1f\t%.1f\t%.2f\t%.2f\t%.3f\n",
			gXc,gYc,
			(float)gX,(float)gY, 
			off.x, off.y,
			confidence);
	}
	fclose(offsets);
	outRast.writeNoThrow("marked.ppm");
	return 0;
}

