/*
A GLUT program using OpenGL, GLUI, and GLSL.

Orion Sky Lawlor, olawlor@acm.org, 2007/01/18 (Public Domain)
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include <GL/glew.h> /* GL Extensions Wrangler, http://glew.sourceforge.net/ */
#include <GL/glut.h> /* GL Utilities Toolkit, http://freeglut.sourceforge.net/ */
#include <GL/glui.h> /* GL User Interface, http://glui.sourceforge.net/ */
#include "osl/vec4.h" /* GLSL-style C++ types, by Orion Lawlor */
#include "osl/mat4.h"
using osl::mat4;
#include "osl/osl.h"

/* Dr. Lawlor's GLSL utilities */
#include "ogl/program.h"
#include "ogl/program.cpp"
#include "ogl/util.cpp"
#include "ogl/fast_mipmaps.c"

#include "ltga.h" /* Lev Povalahev's targa-reading library; http://www.levp.de/3d/files/ltga.zip */
#include "ltga.cpp"

#include "osl/vector2d.h"
using osl::Vector2d;


/* Variables updated by the GUI */
float distance=0.5; /* origin-to-camera distance */
int win_w, win_h; /**< Width and height (in pixels) of our window at last resize */

float eye_dist=0.5; /* distance between left and right cameras */
float center_dist=2.0; /* distance to center of scene, shifted to zero parallax */

/* These are rotation matrices, updated by the GLUI_Rotation spheres */
mat4 modelMat(1.0), viewMat(1.0);

/// Scale by this offset
inline void glScalefv(const float *s) {
	glScalef(s[0],s[1],s[2]);
}
/// Translate by this offset
inline void glTranslatefv(const float *s) {
	glTranslatef(s[0],s[1],s[2]);
}
/** Transpose this mat4.  If the mat4 was a rotation matrix, 
  this actually inverts the matrix. */
mat4 transpose(const mat4 &m) {
	return mat4(
		m[0][0], m[1][0], m[2][0], m[3][0],
		m[0][1], m[1][1], m[2][1], m[3][1],
		m[0][2], m[1][2], m[2][2], m[3][2],
		m[0][3], m[1][3], m[2][3], m[3][3]);
}


/** Read a texture object from this file */
oglTexture *readTextureTGA(const char *fName,
	//double *scale_x,double *scale_y,
	int *file_x,int *file_y) 
{
	LTGA f(fName);
	
	/* Careful!  Not all .tga images have an alpha channel! */
	GLenum rgbOrder=GL_RGBA;
	if (f.GetImageType()==itRGB) rgbOrder=GL_RGB;
	
	/* Scale image up to a power of two size */
	int fw=f.GetImageWidth(), fh=f.GetImageHeight(); /* file size */
	int pw=osl::roundUp2(fw), ph=osl::roundUp2(fh); /* padded, destination size */
	*file_x=(int)fw; *file_y=(int)fh;
	//*scale_x=(double)fw/pw;
	//*scale_y=(double)fh/ph;
	int pixelSize=(rgbOrder==GL_RGBA)?4:3;
	const byte *src=(byte *)f.GetPixels(); /* pixelSize*fw*fh bytes */
	byte *dest=new byte[pixelSize*pw*ph];
	int x,y,c;
	
	/* Loop over the *output* image */
	for (y=0;y<ph;y++) 
	for (x=0;x<pw;x++) {
		double sy=y*(1.0/ph*fh);
		double sx=x*(1.0/pw*fw);
		for (c=0;c<pixelSize;c++) {
			byte s=0;
			if (sx<fw && sy<fh) /* read from source image */
			{
				/* Bilinear interpolation */
				int ix=(int)sx, iy=(int)sy;  /* round down */
				double dx=sx-ix, dy=sy-iy; /* sub-pixel loc. */
				byte tl = src[(iy*fw+ix)*pixelSize+c];
				byte tr = src[(iy*fw+ix+1)*pixelSize+c];
				byte bl = src[((iy+1)*fw+ix)*pixelSize+c];
				byte br = src[((iy+1)*fw+ix+1)*pixelSize+c];
				
				double t = tl + dx*(tr-tl);
				double b = bl + dx*(br-bl);
				s = (byte)(t+dy*(b-t));
				/*
				// Equivalent version with more arithmetic:
				s = tl*(1-dx)*(1-dy)+
				    tr*(dx)*(1-dy)+
				    bl*(1-dx)*(dy)+
				    br*dx*dy;
				*/
			}
			dest[(y*pw+x)*pixelSize+c] = s;
		}
	}
	
	oglTexture *t=new oglTexture(dest, /* pixel data */
			pw,ph, /* width and height */
			oglTexture_mipmap, /* do build mipmaps */
			rgbOrder
			);
	printf("After uploading texture: file size %d x %d\n",
		fw,fh);
	delete[] dest;
	return t;
}

/**
 A stereo-reconstructed 3D object.
 Contains both the texture image, and the geometry.
*/
class stereoObject {
public:
	oglTexture *tex; /* texture image to stretch across geometry */
	int w,h; /* size of texture image, pixels */
	double tx,ty; /* scale factors on X and Y tex coords */
	
	class offset {
	public:
		float px,py; /* texture coordinates of our offset */
		float ox,oy; /* texture coordinate shift for this offset */
	};
	
	/* 2D grid of image-to-image offsets */
	std::vector<std::vector<offset> > off;
	
	stereoObject(std::string pathToDir) {
		const char *stereoOffsets=(pathToDir+"/offsets.txt").c_str();
		FILE *offsets=fopen(stereoOffsets,"r");
		if (offsets==0) oglExit("reading stereo offsets file",stereoOffsets);
		
		tex=readTextureTGA((pathToDir+"/R.tga").c_str(),&w,&h);
		
		int gx,gy;
		offset o;
		//float px,py,ox,oy; /* pixel in texture, and stereo offset */
		float conf; /* confidence */
		while (7==fscanf(offsets,"%d%d%f%f%f%f%f",
			&gx,&gy,&o.px,&o.py,&o.ox,&o.oy,&conf)) {
			if (gx==0) /* start of new row of offsets */
				off.push_back(std::vector<offset>());
			o.px*=1.0/w; o.py*=1.0/h;  /* pixel to tex coords */
			o.py=1.0-o.py; /* flip Y axis upside down (for OpenGL) */
			o.ox*=1.0/w; o.oy*=1.0/h;  /* pixel to tex coords */
			off[gy].push_back(o);
		}
		fclose(offsets);
	}
	
	/* Draw the vertex reconstructed from this image-image offset */
	void drawOffset(const offset &o) {
		/* compute Z depth, based on similar triangles to center distance */
		float z=eye_dist / (o.ox + eye_dist/center_dist);
		//z=10.0*o.ox; // <- cheap hack 
		
		/* Convert from camera-relative Z depth to 3D world coordinates */
		vec3 camera(0,0,0);
		double hFov=tan(45*M_PI/180*0.5); /* tangent of half the horizontal field-of-view */
		double vFov=hFov*1200/1600; /* tangent of half the vertical field-of-view */
		vec3 pixel(hFov*(o.px-0.5),vFov*(o.py-0.5),-1); /* location onscreen */
		vec3 world=z*(pixel-camera)+camera;
		
		glTexCoord2f(o.px,o.py); 
		glVertex3fv(world);
	}
	/* Draw all our geometry */
	void draw(void) {
		tex->bind();
		
		glTranslatef(0,0,center_dist); /* center image */
		glEnable(GL_TEXTURE_2D);
		glBegin(GL_QUADS);
		for (unsigned int y=0;y<off.size()-1;y++)
		for (unsigned int x=0;x<off[y].size()-1;x++)
		{ /* draw offset vertices */
			drawOffset(off[y][x]);
			drawOffset(off[y][x+1]);
			drawOffset(off[y+1][x+1]);
			drawOffset(off[y+1][x]);
		}
		glEnd();
	}
};
stereoObject *so=0;


/* GLUT calls this routine every time our window needs to be redrawn */
void main_display(void) {
	// Clear the color and depth buffers
	glClearColor(0.2,0.4,0.8,0.0); /* sky blue */
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); /* actually clear color and depth */
	
	/* Depth testing */
	glEnable(GL_DEPTH_TEST);
	
	/* Alpha blending */
	glEnable(GL_BLEND);
	           /*  src color  + dest color  */
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	
	// Load all needed matrices into OpenGL
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity(); /*<- clean out any old leftover matrices */
	gluPerspective(90.0, /* <- Y camera field-of-view, in degrees */
		win_w/(float)win_h, /* viewport's aspect ratio: always width over height */
		0.01, /* Near clipping plane depth */
		100.0 /* Far clipping plane depth */
	);
	glTranslatef(0,0,-distance); /* push world origin down the *unrotated* Z axis */
	glMultMatrixf(&viewMat[0][0]); /* camera orientation: could also go under GL_MODELVIEW */
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	
	so->draw();
	
	
	// Display the stuff drawn so far to the screen:
	glutSwapBuffers();
}

/* GLUT calls this routine every time our window changes size */
void main_reshape(int w, int h) {
	win_w=w; win_h=h;
	glViewport(0, 0, (GLsizei)win_w, (GLsizei)(win_h));
}

int main(int argc, char** argv)
{
	glutInitWindowSize(800,600);
	glutInit(&argc,argv);

	// Ask for a back buffer, color buffer, and Z buffer in a window.
	int mode=GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH;
	glutInitDisplayMode (mode);
	int main_window = glutCreateWindow("GLSL Stereo Viewer");

	// Turn on OpenGL extensions, like programmable shaders
	//   WARNING: you MUST call glutCreateWindow before doing this!
	//   If you don't call glewInit, or call it wrong, you'll segfault!
	glewInit();
	
	so=new stereoObject(argv[1]?argv[1]:"src/");
	
	// Ask GLUT to call us when we need to redraw
	glutDisplayFunc (main_display);
	glutReshapeFunc (main_reshape);

	if (1) {
		// Make some sliders (this is the only GLUI code you'll need!)
		GLUI *glui = GLUI_Master.create_glui_subwindow(main_window, 
			GLUI_SUBWINDOW_TOP);
		/* This causes the main window to be redrawn when a GLUI control changes */
		glui->set_main_gfx_window( main_window );

		GLUI_Spinner *spin_zoom = new GLUI_Spinner( glui, "View Distance:", &distance);
		spin_zoom->set_float_limits(0.0, 10.0);
		(new GLUI_Spinner( glui, "Eye Distance:", &eye_dist))
			->set_float_limits(0.0, 1.0);
		(new GLUI_Spinner( glui, "Center Distance:", &center_dist))
			->set_float_limits(0.0, 10.0);
		
		new GLUI_Column(glui);
		//new GLUI_Rotation(glui,"Model:",&modelMat[0][0]);
		//new GLUI_Column(glui,false);
		new GLUI_Rotation(glui,"View:",&viewMat[0][0]);
	}
	
	// Process events until user closes our window...
	glutMainLoop();
	
	return 0;
}
